#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <string.h>
#include <errno.h>
#include <srr.h>

void* thread_function( void *ptr );

typedef struct xmsg
{
	char		*str;
	int			ppid;
} xmsg;

int main(int argc, char** argv)
{
     pthread_t		thread1, thread2;
     char 			*message1 = "Thread 1";
     char 			*message2 = "Thread 2";
     int  			iret1, iret2;
	 xmsg			msg1, msg2;
	 pid_t			pid;
	 char			buf[512];
	 int			 i;
	 size_t			size;
	 
	 msg1.str = message1;
	 msg2.str = message2;
	 msg1.ppid = msg2.ppid = getpid();

	 /* We cannot predict the order in which threads will run.  We
		have arranged the program such that child threads Send() to
		the main thread, so the main thread must have already
		registered with the SRRIPC module.  We force this registration
		with SrrReg().  Child threads will register as necessary, as
		they have no ordering dependencies. */
	 if(SrrReg() == -1) {
			 printf("SrrReg() failed: [%d] %s\n", errno, strerror(errno));
			 exit(1);
	 }
	 
	 /* Create independant threads each of which will execute function */

     printf("Parent process: pid: %d, tid: 0x%lx\n", getpid(), pthread_self());
     iret1 = pthread_create( &thread1, NULL, (void*)&thread_function,
							 (void*) &msg1);
     iret2 = pthread_create( &thread2, NULL, (void*)&thread_function,
							 (void*) &msg2);

	 /* Loop collecting a message from each thread, reply to the message,
		then collect the thread exit status and return. */
	 for (i=0; i<2; i++)
	 {
		 size = sizeof(buf);
		 pid = SrrReceive (0, buf, &size);
		 if (pid != -1)
		 {
			 buf[size] = 0;
			 printf ("Parent received from PID: %d  msg: %s\n", pid, buf);
			 sprintf (buf, "Thanks, %d", pid);
			 Reply (pid, buf, strlen(buf)+1);
		 }
		 else
		 {
			 printf ("Error in Receive: %d: %s\n", errno, strerror(errno));
		 }
	 }

     /* Wait till threads are complete before main continues. Unless we  */
     /* wait we run the risk of executing an exit which will terminate   */
     /* the process and all threads before the threads have completed.   */

     pthread_join( thread1, (void**)&iret1);
     pthread_join( thread2, (void**)&iret2);

     printf("Thread 1 returns: %d\n",iret1);
     printf("Thread 2 returns: %d\n",iret2);
     return(0);
}

void* thread_function (void *ptr)
{
	char		buf[64];
	
	xmsg	*msg = (xmsg*)ptr;
	printf("Thread started: pid: %d, tid: 0x%lx: %s \n",
		   getpid(), pthread_self(), msg->str);
	Send (msg->ppid, msg->str, buf, strlen(msg->str), sizeof(buf));
	printf("Child %d received response to '%s': %s\n",
		   getpid(), msg->str, buf);
	return ((void*)getpid());
}
