#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <string.h>
#include <errno.h>
#include <srr.h>

#define MAXBUF		6400
#define MINBUF		1200
#define NTHREAD		NThreads
#define MAXCOUNT	NMessage

#define MXDEBUG(fmt, args...);
//#define MXDEBUG(fmt, args...) \
//  fprintf(stderr, fmt, ##args);

int NMessage = 10;
int NThreads = 20;
int UseMalloc = 0;
int UseDummyThreads = 0;
int DummySize = 0;

int UsingMX = 0;
int Exiting = 0;
int Waiting = 0;

void* send_thread_function (void *ptr);
void* receive_thread_function (void *ptr);
void* dummy_thread_parent (void *ptr);
void* dummy_thread_child (void *ptr);

char	recvmsgs[20][MAXBUF];
char	inmsgs[20][MAXBUF];
char	outmsgs[20][MAXBUF];

typedef struct tdata
{
	char		fill;
	int			indx;
	pid_t		parent;
} tdata;

pid_t			*rpid;
pid_t			*spid;

int				quiet = 0;

static void help (char* name)
{
printf (
"Pthread test program.\n"
"Usage: %s [-a bytes] [-h] [-q] [-n threads] [-m msgs] [-M] [-p] [-w] [-x]\n"
"\t-a    - allocate this much memory in each receiving thread, and attempt\n"
"\t          to access it every 1-20 millisecnds\n"
"\t-h    - print this helpful message\n"
"\t-q    - do not print out progress information\n"
"\t-n    - start this many receive threads, and this many send threads\n"
"\t-m    - transmit up to this many messages per sending thread\n"
"\t-M    - use malloc() instead of static allocation for buffers\n"
"\t          allows you to set more than 20 threads with the -n option\n"
"\t-p    - constantly create and destroy processes in the background\n"
"\t-w    - introduce a small random wait after receiving a message\n"
"\t-x    - use MX versions of Send and Receive functions\n"
, name);
 exit(-1);
}

int main(int argc, char** argv)
{
     int  			iret;
	 int			i;
	 pthread_t		*senders;
	 pthread_t		*receivers;
	 pthread_t		tid, dummytid;
	 tdata			*data;

	 for (i=1; i<argc; i++)
	 {
		 if (argv[i][0] == '-')
		 {
			 switch (argv[i][1])
			 {
			   case 'a':
				 DummySize = atoi(argv[++i]);
				 break;
			   case 'q':
				 quiet = 1;
				 break;
			   case 'n':
				 NThreads = atoi(argv[++i]);
				 if (!UseMalloc && NThreads > 20)
					 NThreads = 20;
				 break;
			   case 'M':
				 UseMalloc = 1;
				 break;
			   case 'm':
				 NMessage = atoi(argv[++i]);
				 break;
			   case 'p':
				 UseDummyThreads = 1;
				 break;
			   case 'x':
				 UsingMX = 1;
				 break;
			   case 'w':
				 Waiting = 1;
				 break;
			   case 'h':
			   default:
				 help(argv[0]);
				 break;
			 }
		 }
	 }

	 /* We cannot predict the order in which threads will run.  We
		have arranged the program such that child threads Send() to
		the main thread, so the main thread must have already
		registered with the SRRIPC module.  We force this registration
		with SrrReg().  Child threads will register as necessary, as
		they have no ordering dependencies. */
	 if(SrrReg() == -1) {
			 printf("SrrReg() failed: [%d] %s\n", errno, strerror(errno));
			 exit(1);
	 }
	 
	 /* Create independant threads each of which will execute function */

	 rpid = (pid_t*)malloc(NTHREAD * sizeof(pid_t));
	 spid = (pid_t*)malloc(NTHREAD * sizeof(pid_t));
	 senders = (pthread_t*)malloc(NTHREAD * sizeof(pthread_t));
	 receivers = (pthread_t*)malloc(NTHREAD * sizeof(pthread_t));
	 memset (senders, 0, sizeof(pthread_t) * NTHREAD);
	 memset (receivers, 0, sizeof(pthread_t) * NTHREAD);
	 memset (spid, 0, sizeof(pid_t) * NTHREAD);
	 memset (rpid, 0, sizeof(pid_t) * NTHREAD);

	 if (UseDummyThreads)
	 {
		 iret = pthread_create(&dummytid, NULL,
							   (void*)&dummy_thread_parent, NULL);
		 if (iret == -1)
		 {
			 printf ("Failed to create dummy thread parent: %s\n",
					 strerror(errno));
		 }
	 }
	 
	 for (i=0; i<NTHREAD; i++)
	 {
		 data = (tdata*) malloc (sizeof(tdata));
		 data->parent = getpid();
		 data->fill = 0;
		 data->indx = i;
		 
		 iret = pthread_create(&tid, NULL, (void*)&receive_thread_function,
							   (void*) data);
		 if (iret == -1)
		 {
			 printf ("Failed to create receiver: %s\n", strerror(errno));
		 }
		 else
		 {
			 receivers[i] = tid;
		 }
	 }

	 for (i=0; i<NTHREAD; i++)
	 {
		 data = (tdata*) malloc (sizeof(tdata));
		 data->parent = getpid();
		 data->fill = 'a' + i;
		 data->indx = i;
		 
		 iret = pthread_create(&tid, NULL, (void*)&send_thread_function,
							   (void*) data);
		 if (iret == -1)
		 {
			 printf ("Failed to create receiver: %s\n", strerror(errno));
		 }
		 else
		 {
			 senders[i] = tid;
		 }
	 }

     /* Wait till threads are complete before main continues. Unless we  */
     /* wait we run the risk of executing an exit which will terminate   */
     /* the process and all threads before the threads have completed.   */

	 if (!quiet)
		 printf ("Waiting for senders to complete ...\n");
	 
	 for (i=0; i<NTHREAD; i++)
	 {
		 if (senders[i])
		 {
			 if (pthread_join (senders[i], (void**)&iret) == -1)
				 printf ("Could not join tid %lu: %s\n",
						 senders[i], strerror(errno));
		 }
		 else
		 {
			 printf ("No sender index %d\n", i);
		 }
	 }

	 Exiting = 1;

	 if (!quiet)
		 printf ("Cancelling receivers ...\n");
	 
	 for (i=0; i<NTHREAD; i++)
	 {
		 if (receivers[i])
		 {
			 if ((iret = pthread_cancel (receivers[i])) == 0)
			 {
				 pthread_join (receivers[i], (void**)&iret);
				 //printf ("Receiver returned %d\n", iret);
			 }
			 else
			 {
				 printf ("Failed to cancel receiver %d: %d: %s\n",
						 rpid[i], iret, strerror(iret));
			 }
		 }
	 }

	 if (UseDummyThreads)
	 {
		 if (!quiet)
			 printf ("Cancelling dummy parent ...\n");
		 pthread_join (dummytid, (void**)&iret);
		 if (!quiet)
			 printf ("   after %d child threads\n", iret);
	 }
	 
     return(0);
}

void small_sleep()
{
	if (Waiting)
	{
		struct timespec req, rem;
		int		rval;
				
		req.tv_sec = 0;
		req.tv_nsec = ((random() % 20)+1) * 1000000;
		rval = nanosleep (&req, &rem);
		if (rval == -1)
		{
			printf ("nanosleep: %s\n", strerror(errno));
		}
	}
}

void* dummy_thread_parent (void *ptr)
{
	int			iret, count=0;
	pthread_t	tid;
	
	for (;!Exiting;)
	{
		iret = pthread_create(&tid, NULL, (void*)&dummy_thread_child, NULL);
		if (iret == -1)
		{
			printf ("Failed to create dummy child %d: %s\n",
					count, strerror(errno));
		}
		pthread_join (tid, NULL);
		count++;
	}
	return ((void*)count);
}

void* dummy_thread_child (void* ptr)
{
	char			buf[32];
	unsigned int	copies;

	sprintf (buf, "dummy%x", SrrGetpid());
	SrrReg();
	SrrNameAttach (0, buf);
	small_sleep();
	SrrNameLocate (0, buf, 0, &copies);
	return (NULL);
}

void* receive_thread_function (void *ptr)
{
	char		*in;
	tdata		*data = (tdata*)ptr;
	int			count, maxcount, len, i;
	pid_t		sender, timer_proxy;
	srr_timer_t	timer;
	struct itimerspec	it;
	struct sigevent		ev;
	int tick = 0;
	char		*membuf;		// Used to consume memory and cause swap

	if (DummySize)
		membuf = (char*) malloc (DummySize);
	
	rpid[data->indx] = SrrGetpid();
	if (errno)
		printf ("Receiver: SrrGetpid error: %s\n", strerror(errno));

	pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
	
	if (UseMalloc)
		in = (char*)malloc (MAXBUF);
	else
		in = recvmsgs[data->indx];

	memset (in, data->fill, MAXBUF);
	maxcount = random() % MAXCOUNT;

	timer_proxy = SrrProxyAttach (0, NULL, 0, 0);
	ev.sigev_signo = -timer_proxy;
	timer = SrrTimerCreate (CLOCK_REALTIME, &ev);
	it.it_interval.tv_sec = 0;
	it.it_interval.tv_nsec = 100000;
	it.it_value.tv_sec = 0;
	it.it_value.tv_nsec = 100000;
	SrrTimerSettime (timer, 0, &it, 0);
	
	for (count = 0; ; count++)
	{
		len = MAXBUF;
		if (UsingMX)
		{
			struct _mxfer_entry msgmx[2];
			_setmx (&msgmx[0], in, len/2);
			_setmx (&msgmx[1], &((char*)in)[msgmx[0].mxfer_len],
					len - msgmx[0].mxfer_len);
			sender = SrrReceivemx (0, 2, msgmx);
			len = msgmx[0].mxfer_len + msgmx[1].mxfer_len;
		}
		else
		{
			sender = SrrReceive (0, in, &len);
		}
		
		if (sender == -1)
		{
			if (Exiting == 0)
				printf ("Receiver: %d: count: %d: tick %d: SrrReceive error: %s\n",
						SrrGetpid(), count, tick, strerror(errno));
			else
				printf ("Receiver: %d: count: %d: tick %d: controlled exit.\n",
						SrrGetpid(), count, tick);
			break;
		}
		else if (sender == timer_proxy)
		{
			int	j;
			
			tick++;
			if (DummySize)
			{
				for (j=0; j<DummySize; j+=4000)
					membuf[j] = tick;
			}
		}
		else
		{
			if (!quiet)
			{
				printf ("R(p:%d, l:%d: s:%d, tick: %d)\n",
						SrrGetpid(), len, sender, tick);
				fflush (stdout);
			}
			
			for (i=0; i<len; i++)
			{
				if (in[i] != in[0])
				{
				//	printf ("Receiver: buffer incomplete at char %d (char %d) of %d bytes\n",
				//			i, in[i], len);
					break;
				}
			}

			small_sleep();

			if (SrrReply (sender, in, len) == -1)
			{
				printf ("Receiver: %d: SrrReply error: %s\n",
						SrrGetpid(), strerror(errno));
			}
		}
	}
	rpid[data->indx] = 0;
	return ((void*)SrrGetpid());
}

void* send_thread_function (void *ptr)
{
	char		*out, *in;
	tdata		*data = (tdata*)ptr;
	int			count, maxcount, len, rec, i, misscount=0, inlen;

	spid[data->indx] = SrrGetpid();
	
	if (UseMalloc)
	{
		out = (char*)malloc (MAXBUF);
		in = (char*)malloc (MAXBUF);
	}
	else
	{
		out = outmsgs[data->indx];
		in = inmsgs[data->indx];
	}
	
	memset (out, data->fill, MAXBUF);
	maxcount = random() % MAXCOUNT;

	for (count = 0; count < maxcount; count++)
	{
		len = random() % (MAXBUF-MINBUF) + MINBUF;
		rec = random() % NTHREAD;
		if (rpid[rec] == 0)
		{
			for (i=rec; i<NTHREAD; i++)
			{
				if (rpid[i] != 0)
				{
					rec = i;
					break;
				}
			}
			if (rpid[rec] == 0)
			{
				for (i=0; i<rec; i++)
				{
					if (rpid[i] != 0)
					{
						rec = i;
						break;
					}
				}
			}

			if (rpid[rec] == 0)
				return ((void*)SrrGetpid());
		}

		if (!quiet)
		{
			printf ("S(p:%d, l:%d, r:%d)\n", SrrGetpid(), len, rpid[rec]);
			fflush (stdout);
		}

		if (UsingMX)
		{
			struct _mxfer_entry smsgmx[2];
			struct _mxfer_entry rmsgmx[2];

			_setmx(&smsgmx[0], out, len/2);
			_setmx(&smsgmx[1], &((char*)out)[smsgmx[0].mxfer_len],
				   len - smsgmx[0].mxfer_len);

			_setmx(&rmsgmx[0], in, MAXBUF/2);
			_setmx(&rmsgmx[1], &((char*)in)[rmsgmx[0].mxfer_len],
				   MAXBUF - rmsgmx[0].mxfer_len);
      MXDEBUG("Size of send mxfer entries:  1: %d  2: %d\n", sizeof(smsgmx[0]), sizeof(smsgmx[1]));		
      MXDEBUG("Size of reply mxfer entries: 1: %d  2: %d\n", sizeof(rmsgmx[0]), sizeof(rmsgmx[1]));
			inlen = SrrSendmx(rpid[rec], 2, 2, smsgmx, rmsgmx);

			if (rmsgmx[0].mxfer_len < MAXBUF/2 &&
				rmsgmx[1].mxfer_len != 0)
			{
				printf ("Mxfer length error: len %d, not %d + %d\n",
						len, rmsgmx[0].mxfer_len, rmsgmx[1].mxfer_len);
			}
		}
		else
		{
			inlen = SrrSend (rpid[rec], out, in, len, MAXBUF);
		}
		
		if (inlen == -1)
		{
			printf ("Sender: %d: SrrSend error: %s\n",
					SrrGetpid(), strerror(errno));
		}
		else if (inlen != len)
		{
	//		printf ("Sender: %d: SrrSend different size: sent %d, got %d\n",
		//			SrrGetpid(), len, inlen);
		}
		else
		{
			if (memcmp (out, in, len) != 0)
			{
				misscount++;
				for (i=0; i<len; i++)
				{
					if (in[i] != out[i])
					{
						break;
					}
				}
				printf ("Sender %d, msg %d, len %d, char %c, buffer mismatch at char %d: out %d, in %d.\n",
						SrrGetpid(), count, len, data->fill, i, out[i], in[i]);
				printf ("  Missed %d of %d.  out: %p(%d), in: %p(%d)\n",
						misscount, count+1,
						out, len,
						in, inlen
						);
				for (i=0; i<MAXBUF; i++)
				{
					if (out[i] != data->fill)
					{
						printf ("  out buffer corrupt at %d: char %d\n",
								i, out[i]);
					}
				}
			}
			small_sleep();
		}
	}
	spid[data->indx] = 0;
	return ((void*)SrrGetpid());
}
