/*
 * reply.c: simple program that receives and replies to a message
 *
 * $Id: reply.c 132 2011-07-01 14:35:01Z meaton $
 */

#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "srr.h"

int main(int argc, char* argv[])
{
	int		opt;
	pid_t	pid = 0;
	pid_t	client = 0;
	char*	name = 0;
	char	recv[BUFSIZ] = "";
	char	reply[BUFSIZ] = "";
	int i,mlen,rlen;

	sprintf(reply, "reply from %d", getpid());

	while((opt = getopt(argc, argv, "p:n:")) != -1)
	{
		switch(opt)
		{
		case 'p':
			pid = atoi(optarg);
			break;

		case 'n':
			name = optarg;
			break;

		default:
			exit(1);
		}
	}

	if(SrrReg() == -1)
	{
		printf("reply %d: InitSrr() failed: [%d] %s\n",
			getpid(), errno, strerror(errno));
		exit(1);
	}

	if(name)
	{
		if(qnx_name_attach(0, name) == -1)
		{
			printf("reply %d: qnx_name_attach(%s) failed: [%d] %s\n",
				getpid(), name, errno, strerror(errno));
			exit(1);
		}
	}

	if((client = Receive(pid, recv, BUFSIZ)) == -1)
	{
		printf("reply %d: Receive(%d) failed: [%d] %s\n",
			getpid(), pid, errno, strerror(errno));
		exit(1);
	}

	printf("reply %d: received: \"%s\"\n", getpid(), recv);
	
	mlen = rlen = strlen(recv);
	for (i = 0; i < mlen+1; ++i)
	{
		memset(recv,0,rlen);
		rlen = Readmsg(client, i, recv, mlen/3);
		recv[rlen] = '\0';
		printf("readmsg %d: read %d bytes: %s           \r", getpid(),rlen,recv);
		fflush(stdout);
		sleep(1);
	}
	printf("\n");
	
	if(Reply(client, reply, strlen(reply)+1) == -1)
	{
		printf("reply %d: Reply(%d) failed: [%d] %s\n",
			getpid(), client, errno, strerror(errno));
		exit(1);
	}
	
	printf("reply %d: done\n", getpid());

	return 0;
}

