/*
 * User code for constructing opcodes for use in SRRIPC module
 * interrupt handler.
 */

/* SBC_MAPA - map an address buffer (physical memory) */
/* SBC_MAPS - map a storage buffer (allocated memory) */

#include	<sys/types.h>
#include	<time.h>
#include	<string.h>
#include	<errno.h>
#include	<sys/ioctl.h>
#include	<inttypes.h>
#include	"../mod/srr_bytecode.h"
#include	"../inc/srr.h"
#include	"../inc/srr_if.h"

int Srr_Mkop_mapa (char* buf, uint8_t adbuf, uint32_t offset, uint32_t size)
{
	int		i = 0;
	
	i += SBPUT8(SBC_MAPA, buf, i);
	i += SBPUT8(adbuf, buf, i);
	i += SBPUT32(offset, buf, i);
	i += SBPUT32(size, buf, i);
	return (i);
}
 
int Srr_Mkop_maps(char* buf, uint8_t stbuf, uint32_t size)
{
	int		i = 0;
	
	i += SBPUT8(SBC_MAPS, buf, i);
	i += SBPUT8(stbuf, buf, i);
	i += SBPUT32(size, buf, i);
	return (i);
}

int Srr_Mkop_mapp(char* buf, uint16_t ioport, uint16_t size)
{
	int		i = 0;
	
	i += SBPUT8(SBC_MAPP, buf, i);
	i += SBPUT16(ioport, buf, i);
	i += SBPUT16(size, buf, i);
	return (i);
}

/* SBC_LDREG - load a register with an immediate value */

int Srr_Mkop_ldreg(char* buf, uint8_t reg, uint32_t value)
{
	int		i = 0;
	
	i += SBPUT8(SBC_LDREG, buf, i);
	i += SBPUT8(reg, buf, i);
	i += SBPUT32(value, buf, i);
	return (i);
}

/* SBC_LDREGA - load a register with a byte/word/long from an address buffer */
/* SBC_LDREGS - load a register with a byte/word/long from a storage buffer */

int Srr_Mkop_ldregas(char* buf, uint8_t opcode, uint8_t reg,
					 uint8_t bufno, uint32_t offset, uint8_t size)
{
	int		i = 0;
	
	i += SBPUT8(opcode, buf, i);
	i += SBPUT8(reg, buf, i);
	i += SBPUT8(bufno, buf, i);
	i += SBPUT32(offset, buf, i);
	i += SBPUT8(size, buf, i);
	return (i);
}

int Srr_Mkop_ldrega (char* buf, uint8_t reg, uint8_t bufno,
					 uint32_t offset, uint8_t size)
{
	return (Srr_Mkop_ldregas(buf, SBC_LDREGA, reg, bufno, offset, size));
}

int Srr_Mkop_ldregs (char* buf, uint8_t reg, uint8_t bufno,
					 uint32_t offset, uint8_t size)
{
	return (Srr_Mkop_ldregas(buf, SBC_LDREGS, reg, bufno, offset, size));
}

/* SBC_LDREGP - load a register with a byte/word/long from an I/O port.
                The address of the I/O port is in the port_reg */

int Srr_Mkop_ldregp (char* buf, uint8_t reg, uint8_t port_reg, uint8_t size)
{
	int		i = 0;
	
	i += SBPUT8(SBC_LDREGP, buf, i);
	i += SBPUT8(reg, buf, i);
	i += SBPUT8(port_reg, buf, i);
	i += SBPUT8(size, buf, i);
	return (i);
}

/* SBC_STREGA - store a 1/2/4 byte value from a register to address buffer */
/* SBC_STREGS - store a 1/2/4 byte value from a register to storage buffer */

int Srr_Mkop_stregas (char* buf, uint8_t opcode, uint8_t reg, uint8_t bufno,
					  uint32_t offset, uint8_t size)
{
	int		i = 0;
	
	i += SBPUT8(opcode, buf, i);
	i += SBPUT8(reg, buf, i);
	i += SBPUT8(bufno, buf, i);
	i += SBPUT32(offset, buf, i);
	i += SBPUT8(size, buf, i);
	return (i);
}

int Srr_Mkop_strega (char* buf, uint8_t opcode, uint8_t reg, uint8_t bufno,
					 uint32_t offset, uint8_t size)
{
	return (Srr_Mkop_stregas (buf, SBC_STREGA, reg, bufno, offset, size));
}

int Srr_Mkop_stregs (char* buf, uint8_t opcode, uint8_t reg, uint8_t bufno,
					 uint32_t offset, uint8_t size)
{
	return (Srr_Mkop_stregas (buf, SBC_STREGS, reg, bufno, offset, size));
}

int Srr_Mkop_stregp (char* buf, uint8_t reg, uint8_t port_reg, uint8_t size)
{
	int		i = 0;
	
	i += SBPUT8(SBC_STREGP, buf, i);
	i += SBPUT8(reg, buf, i);
	i += SBPUT8(port_reg, buf, i);
	i += SBPUT8(size, buf, i);
	return (i);
}

/* Binary register operations */

int Srr_Mkop_regbin (char* buf, uint8_t opcode, uint8_t reg1, uint8_t reg2)
{
	int		i = 0;
	
	i += SBPUT8(opcode, buf, i);
	i += SBPUT8(reg1, buf, i);
	i += SBPUT8(reg2, buf, i);
	return (i);
}

int Srr_Mkop_add (char* buf, uint8_t reg1, uint8_t reg2)
{
	return (Srr_Mkop_regbin (buf, SBC_ADD, reg1, reg2));
}

int Srr_Mkop_sub (char* buf, uint8_t reg1, uint8_t reg2)
{
	return (Srr_Mkop_regbin (buf, SBC_SUB, reg1, reg2));
}

int Srr_Mkop_mul (char* buf, uint8_t reg1, uint8_t reg2)
{
	return (Srr_Mkop_regbin (buf, SBC_MUL, reg1, reg2));
}

int Srr_Mkop_div (char* buf, uint8_t reg1, uint8_t reg2)
{
	return (Srr_Mkop_regbin (buf, SBC_DIV, reg1, reg2));
}

int Srr_Mkop_and (char* buf, uint8_t reg1, uint8_t reg2)
{
	return (Srr_Mkop_regbin (buf, SBC_AND, reg1, reg2));
}

int Srr_Mkop_xor (char* buf, uint8_t reg1, uint8_t reg2)
{
	return (Srr_Mkop_regbin (buf, SBC_XOR, reg1, reg2));
}

int Srr_Mkop_or (char* buf, uint8_t reg1, uint8_t reg2)
{
	return (Srr_Mkop_regbin (buf, SBC_OR, reg1, reg2));
}

int Srr_Mkop_shl (char* buf, uint8_t reg1, uint8_t reg2)
{
	return (Srr_Mkop_regbin (buf, SBC_SHL, reg1, reg2));
}

int Srr_Mkop_shr (char* buf, uint8_t reg1, uint8_t reg2)
{
	return (Srr_Mkop_regbin (buf, SBC_SHR, reg1, reg2));
}

int Srr_Mkop_cpyr (char* buf, uint8_t reg1, uint8_t reg2)
{
	return (Srr_Mkop_regbin (buf, SBC_CPYR, reg1, reg2));
}

/* Unary register operations */

int Srr_Mkop_regunary (char* buf, uint8_t opcode, uint8_t reg)
{
	int		i = 0;
	
	i += SBPUT8(opcode, buf, i);
	i += SBPUT8(reg, buf, i);
	return (i);
}

int Srr_Mkop_not (char* buf, uint8_t opcode, uint8_t reg)
{
	return (Srr_Mkop_regunary (buf, SBC_NOT, reg));
}

int Srr_Mkop_inc (char* buf, uint8_t opcode, uint8_t reg)
{
	return (Srr_Mkop_regunary (buf, SBC_INC, reg));
}

int Srr_Mkop_dec (char* buf, uint8_t opcode, uint8_t reg)
{
	return (Srr_Mkop_regunary (buf, SBC_DEC, reg));
}

/* SBC_CPYAS - copy from address buffer to storage buffer */
/* SBC_CPYSA - copy from storage buffer to address buffer */

int Srr_Mkop_cpy (char* buf, uint8_t opcode, uint8_t buf1, uint32_t off1,
				  uint8_t buf2, uint32_t off2, uint32_t size)
{
	int		i = 0;
	
	i += SBPUT8(opcode, buf, i);
	i += SBPUT8(buf1, buf, i);
	i += SBPUT32(off1, buf, i);
	i += SBPUT8(buf2, buf, i);
	i += SBPUT32(off2, buf, i);
	i += SBPUT32(size, buf, i);
	return (i);
}

int Srr_Mkop_cpyas (char* buf, uint8_t buf1, uint32_t off1,
					uint8_t buf2, uint32_t off2, uint32_t size)
{
	return (Srr_Mkop_cpy (buf, SBC_CPYAS, buf1, off1, buf2, off2, size));
}

int Srr_Mkop_cpysa (char* buf, uint8_t buf1, uint32_t off1,
					uint8_t buf2, uint32_t off2, uint32_t size)
{
	return (Srr_Mkop_cpy (buf, SBC_CPYSA, buf1, off1, buf2, off2, size));
}

/* Miscellaneous */

int Srr_Mkop_proxy (char* buf, pid_t proxy)
{
	int		i = 0;
	
	i += SBPUT8(SBC_PROXY, buf, i);
	i += SBPUT32(proxy, buf, i);
	return (i);
}

int Srr_Mkop_noarg (char* buf, uint8_t opcode)
{
	int		i = 0;
	
	i += SBPUT8(opcode, buf, i);
	return (i);
}

int Srr_Mkop_bytearg (char* buf, uint8_t opcode, uint8_t arg)
{
	int		i = 0;
	
	i += SBPUT8(opcode, buf, i);
	i += SBPUT8(arg, buf, i);
	return (i);
}

int Srr_Mkop_detach (char* buf)
{
	return (Srr_Mkop_noarg (buf, SBC_DETACH));
}

int Srr_Mkop_disable (char* buf)
{
	return (Srr_Mkop_noarg (buf, SBC_DISABLE));
}

int Srr_Mkop_enable (char* buf)
{
	return (Srr_Mkop_noarg (buf, SBC_ENABLE));
}

int Srr_Mkop_handled (char* buf, uint8_t handled)
{
	return (Srr_Mkop_bytearg (buf, SBC_HANDLED, handled));
}

int Srr_Mkop_attach (char* buf, uint8_t irq)
{
	int		i = 0;
	
	i += SBPUT8(SBC_ATTACH, buf, i);
	i += SBPUT8(irq, buf, i);
	return (i);
}

int Srr_Mkop_read (char* buf, int bufno, int offset, int rdlen)
{
	int		i = 0;
	i += SBPUT8(SBC_READ, buf, i);
	i += SBPUT8(bufno, buf, i);
	i += SBPUT32(offset, buf, i);
	i += SBPUT32(rdlen, buf, i);
	return (i);
}

int Srr_Mkop_write (char* buf, int bufno, int offset, int wrlen)
{
	int		i = 0;
	i += SBPUT8(SBC_WRITE, buf, i);
	i += SBPUT8(bufno, buf, i);
	i += SBPUT32(offset, buf, i);
	i += SBPUT32(wrlen, buf, i);
	return (i);
}

/* ------------------ Wrappers on top of wrappers -------------------- */

int SrrBytecodeExt (int immediate, int irqhandle, void* prog, int len,
					void *extbuf, int extlen)
{
	struct SrrIRQArgs	s;
	int					err = -1;

	s.type = (immediate ? SRRIRQ_IMMEDIATE : SRRIRQ_INTERRUPT);
	s.irq_handle = irqhandle;
	s.len = len;
	s.extra = extbuf;
	s.extralen = extlen;
	memcpy (s.buffer, prog, len);


	while(((SrrFd() != -1) || (SrrReg() != -1)) &&
		  ((err = ioctl(SrrFd(), SRR_IOCIRQ, &s)) == -1) &&
          (errno == EBUSY))
	{
		SrrUnreg();
	}

	if (err == -1)
		return (err);
	else
		return (s.irq_handle);
}

int SrrBytecode (int immediate, int irqhandle, void* prog, int len)
{
	return (SrrBytecodeExt (immediate, irqhandle, prog, len, 0, 0));
}

int SrrIrq_Attach (int irq)
{
	char		cmd[SRRIRQ_MAXCMD];
	int			len;
	
	len = Srr_Mkop_attach (cmd, irq);
	return(SrrBytecode (1, 17, cmd, len));
}

static int SrrIrq_Noarg (int irqhandle, int opcode)
{
	char		cmd[SRRIRQ_MAXCMD];
	int			len;
	
	len = Srr_Mkop_noarg (cmd, opcode);
	return(SrrBytecode (1, irqhandle, cmd, len));
}

int SrrIrq_Proxy (int irqhandle, pid_t proxy)
{
	char		cmd[SRRIRQ_MAXCMD];
	int			len;
	
	len = Srr_Mkop_proxy (cmd, proxy);
	return(SrrBytecode (1, irqhandle, cmd, len));
}

int SrrIrq_Mapa (int irqhandle, int bufno, int offset, int size)
{
	char		cmd[SRRIRQ_MAXCMD];
	int			len;
	
	len = Srr_Mkop_mapa (cmd, bufno, offset, size);
	return(SrrBytecode (1, irqhandle, cmd, len));
}

int SrrIrq_Maps (int irqhandle, int bufno, int size)
{
	char		cmd[SRRIRQ_MAXCMD];
	int			len;
	
	len = Srr_Mkop_maps (cmd, bufno, size);
	return(SrrBytecode (1, irqhandle, cmd, len));
}

int SrrIrq_Mapp (int irqhandle, int ioport, int size)
{
	char		cmd[SRRIRQ_MAXCMD];
	int			len;

	len = Srr_Mkop_mapp (cmd, ioport, size);
	return(SrrBytecode (1, irqhandle, cmd, len));
}

int SrrIrq_Ldreg (int irqhandle, int immed, int reg, int value)
{
	char		cmd[SRRIRQ_MAXCMD];
	int			len;
	
	len = Srr_Mkop_ldreg (cmd, reg, value);
	return(SrrBytecode (immed, irqhandle, cmd, len));
}

static int SrrIrq_Ldregas (int irqhandle, int opcode, int immed, int reg,
						   int bufno, int offset, int size)
{
	char		cmd[SRRIRQ_MAXCMD];
	int			len;
	
	len = Srr_Mkop_ldregas (cmd, opcode, reg, bufno, offset, size);
	return(SrrBytecode (immed, irqhandle, cmd, len));
}

int SrrIrq_Ldrega (int irqhandle, int immed, int reg,
				   int bufno, int offset, int size)
{
	return (SrrIrq_Ldregas (irqhandle, SBC_LDREGA, immed,
							reg, bufno, offset, size));
}

int SrrIrq_Ldregs (int irqhandle, int immed, int reg,
				   int bufno, int offset, int size)
{
	return (SrrIrq_Ldregas (irqhandle, SBC_LDREGS, immed,
							reg, bufno, offset, size));
}

int SrrIrq_Ldregp (int irqhandle, int immed, int reg,
				   int port_reg, int size)
{
	char		cmd[SRRIRQ_MAXCMD];
	int			len;
	
	len = Srr_Mkop_ldregp (cmd, reg, port_reg, size);
	return(SrrBytecode (immed, irqhandle, cmd, len));
}

static int SrrIrq_Stregas (int irqhandle, int opcode, int immed, int reg,
						   int bufno, int offset, int size)
{
	char		cmd[SRRIRQ_MAXCMD];
	int			len;
	
	len = Srr_Mkop_stregas (cmd, opcode, reg, bufno, offset, size);
	return(SrrBytecode (immed, irqhandle, cmd, len));
}

int SrrIrq_Strega (int irqhandle, int immed, int reg,
				   int bufno, int offset, int size)
{
	return (SrrIrq_Stregas (irqhandle, SBC_STREGA, immed,
							reg, bufno, offset, size));
}

int SrrIrq_Stregs (int irqhandle, int immed, int reg,
				   int bufno, int offset, int size)
{
	return (SrrIrq_Stregas (irqhandle, SBC_STREGS, immed,
							reg, bufno, offset, size));
}

int SrrIrq_Stregp (int irqhandle, int immed, int reg,
				   int port_reg, int size)
{
	char		cmd[SRRIRQ_MAXCMD];
	int			len;
	
	len = Srr_Mkop_stregp (cmd, reg, port_reg, size);
	return(SrrBytecode (immed, irqhandle, cmd, len));
}

static int SrrIrq_Regbin (int irqhandle, int opcode, int immed,
						  int reg1, int reg2)
{
	char		cmd[SRRIRQ_MAXCMD];
	int			len;
	
	len = Srr_Mkop_regbin (cmd, opcode, reg1, reg2);
	return(SrrBytecode (immed, irqhandle, cmd, len));
}

int SrrIrq_Add (int irqhandle, int immed, int reg1, int reg2)
{
	return (SrrIrq_Regbin (irqhandle, SBC_ADD, immed, reg1, reg2));
}

int SrrIrq_Sub (int irqhandle, int immed, int reg1, int reg2)
{
	return (SrrIrq_Regbin (irqhandle, SBC_SUB, immed, reg1, reg2));
}

int SrrIrq_Mul (int irqhandle, int immed, int reg1, int reg2)
{
	return (SrrIrq_Regbin (irqhandle, SBC_MUL, immed, reg1, reg2));
}

int SrrIrq_Div (int irqhandle, int immed, int reg1, int reg2)
{
	return (SrrIrq_Regbin (irqhandle, SBC_DIV, immed, reg1, reg2));
}

int SrrIrq_And (int irqhandle, int immed, int reg1, int reg2)
{
	return (SrrIrq_Regbin (irqhandle, SBC_AND, immed, reg1, reg2));
}

int SrrIrq_Xor (int irqhandle, int immed, int reg1, int reg2)
{
	return (SrrIrq_Regbin (irqhandle, SBC_XOR, immed, reg1, reg2));
}

int SrrIrq_Or (int irqhandle, int immed, int reg1, int reg2)
{
	return (SrrIrq_Regbin (irqhandle, SBC_OR, immed, reg1, reg2));
}

int SrrIrq_Shl (int irqhandle, int immed, int reg1, int reg2)
{
	return (SrrIrq_Regbin (irqhandle, SBC_SHL, immed, reg1, reg2));
}

int SrrIrq_Shr (int irqhandle, int immed, int reg1, int reg2)
{
	return (SrrIrq_Regbin (irqhandle, SBC_SHR, immed, reg1, reg2));
}

int SrrIrq_Cpyr (int irqhandle, int immed, int reg1, int reg2)
{
	return (SrrIrq_Regbin (irqhandle, SBC_CPYR, immed, reg1, reg2));
}

static int SrrIrq_Regunary (int irqhandle, int opcode, int immed, int reg)
{
	char		cmd[SRRIRQ_MAXCMD];
	int			len;
	
	len = Srr_Mkop_regunary (cmd, opcode, reg);
	return(SrrBytecode (immed, irqhandle, cmd, len));
}

int SrrIrq_Not (int irqhandle, int immed, int reg)
{
	return (SrrIrq_Regunary (irqhandle, SBC_NOT, immed, reg));
}

int SrrIrq_Inc (int irqhandle, int immed, int reg)
{
	return (SrrIrq_Regunary (irqhandle, SBC_INC, immed, reg));
}

int SrrIrq_Dec (int irqhandle, int immed, int reg)
{
	return (SrrIrq_Regunary (irqhandle, SBC_DEC, immed, reg));
}

static int SrrIrq_Cpy (int irqhandle, int opcode, int immed, int buf1,
					   int off1, int buf2, int off2, int size)
{
	char		cmd[SRRIRQ_MAXCMD];
	int			len;
	
	len = Srr_Mkop_cpy (cmd, opcode, buf1, off1, buf2, off2, size);
	return(SrrBytecode (immed, irqhandle, cmd, len));
}

int SrrIrq_Cpyas (int irqhandle, int immed, int adbuf, int adoff,
				  int stbuf, int stoff, int size)
{
	return (SrrIrq_Cpy (irqhandle, SBC_CPYAS, immed, adbuf,
						adoff, stbuf, stoff, size));
}

int SrrIrq_Cpysa (int irqhandle, int immed, int stbuf, int stoff,
				  int adbuf, int adoff, int size)
{
	return (SrrIrq_Cpy (irqhandle, SBC_CPYSA, immed, stbuf,
						stoff, adbuf, adoff, size));
}

int SrrIrq_Detach (int irqhandle)
{
	return (SrrIrq_Noarg (irqhandle, SBC_DETACH));
}

int SrrIrq_Disable (int irqhandle)
{
	return (SrrIrq_Noarg (irqhandle, SBC_DISABLE));
}

int SrrIrq_Enable (int irqhandle)
{
	return (SrrIrq_Noarg (irqhandle, SBC_ENABLE));
}

int SrrIrq_Read (int irqhandle, int bufno, int offset, int rdlen, char* rdbuf)
{
	char		cmd[SRRIRQ_MAXCMD];
	int			len;
	
	len = Srr_Mkop_read (cmd, bufno, offset, rdlen);
	return(SrrBytecodeExt (1, irqhandle, cmd, len, rdbuf, rdlen));
}

int SrrIrq_Write (int irqhandle, int bufno, int offset, int rdlen, char* rdbuf)
{
	char		cmd[SRRIRQ_MAXCMD];
	int			len;
	
	len = Srr_Mkop_write (cmd, bufno, offset, rdlen);
	return(SrrBytecodeExt (1, irqhandle, cmd, len, rdbuf, rdlen));
}

int SrrIrq_Handled (int irqhandle, int handled)
{
	char		cmd[SRRIRQ_MAXCMD];
	int			len;
	
	len = Srr_Mkop_handled (cmd, handled);
	return(SrrBytecode (0, irqhandle, cmd, len));
}
